/*-
 * Copyright (C) @BABOLO  2009 May 07 http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C) @BABOLO  2009 May 07 http://www.babolo.ru/\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: bsdfig_in.c,v 1.6 2013/01/25 18:51:36 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <sysexits.h>
#include <stdio.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include <multilar.h>
#include <mife.h>
#include "bsdfig.h"
#include "bsdfig_local.h"
#include "bsdfig_in.h"

#define SET(T, X) do{                                                                                 \
    ((u_int64_t*)(T))[(X) / BSDFIG_BASELEN] |= (1LLU << ((X) % BSDFIG_BASELEN));                      \
} while(0)

static const size_t acoffs[] =
{ offsetof(bsdfig_allconf, my_pipes)
, offsetof(bsdfig_allconf, my_queues)
, offsetof(bsdfig_allconf, my_disables)
, offsetof(bsdfig_allconf, my_ipfwrules)
, offsetof(bsdfig_allconf, my_ipfwtables)
, offsetof(bsdfig_allconf, my_fibs)
};

typedef struct flocal {
    u_int32_t bitlen:8
    ,         secnum:8
    ,         waitq:2
    ;
} flocal;

static int
c_endl(bsdfig_conf *cnf, const u_char **it, flocal flocal) {
    (*it)++;
    return(EX_OK);
}

static int
c_error(bsdfig_conf *cnf, const u_char **it, flocal flocal) {
    ifBLIN_QV1(cnf->flags) warnx("bsdfig_confread: illegal section name");
    return(EX_DATAERR);
}

static int
c_none(bsdfig_conf *cnf, const u_char **it, flocal flocal) {
    return(-1);
}


static int
c_lists(bsdfig_conf *cnf, const u_char **it, flocal flocal) {
    enum classes
    { L0 /* end of string */
    , Lb /* '(' */
    , Le /* ')' */
    , Li /* '[' */
    , Lq /* ']' */
    , Ld /* '1'..'9' */
    , Lh /* 'x' */
    , Lg /* 'a'..'f' */
    , Lz /* '0' */
    , Lp /* '.' */
    , Ls /* ' ' */
    , La /* other */
    , Lx /* last */
    };

    static const u_char recode[256] =
    {  L0, La, La, La,  La, La, La, Ls,  Ls, Ls, L0, L0,  La, L0, La, La
    ,  La, La, La, La,  La, La, La, La,  La, La, La, La,  La, La, La, La
    ,  Ls, La, La, La,  La, La, La, La,  Lb, Le, La, La,  La, La, Lp, La
    ,  Lz, Ld, Ld, Ld,  Ld, Ld, Ld, Ld,  Ld, Ld, La, La,  La, La, La, La

    ,  La, Lg, Lg, Lg,  Lg, Lg, Lg, La,  La, La, La, La,  La, La, La, La
    ,  La, La, La, La,  La, La, La, La,  Lh, La, La, Li,  La, Lq, La, La
    ,  La, Lg, Lg, Lg,  Lg, Lg, Lg, La,  La, La, La, La,  La, La, La, La
    ,  La, La, La, La,  La, La, La, La,  Lh, La, La, La,  La, La, La, La

    ,  La, La, La, La,  La, La, La, La,  La, La, La, La,  La, La, La, La
    ,  La, La, La, La,  La, La, La, La,  La, La, La, La,  La, La, La, La
    ,  La, La, La, La,  La, La, La, La,  La, La, La, La,  La, La, La, La
    ,  La, La, La, La,  La, La, La, La,  La, La, La, La,  La, La, La, La

    ,  La, La, La, La,  La, La, La, La,  La, La, La, La,  La, La, La, La
    ,  La, La, La, La,  La, La, La, La,  La, La, La, La,  La, La, La, La
    ,  La, La, La, La,  La, La, La, La,  La, La, La, La,  La, La, La, La
    ,  La, La, La, La,  La, La, La, La,  La, La, La, La,  La, La, La, La
    };

    enum states
    { Di /* Wait for ( */
    , Da /* Wait for ]( */
    , Dd /* Wait for 0 */
    , De /* Wait for ]0 */
    , Dh /* Wait for x */
    , Dc /* Wait for 9 */
    , Dp /* Wait for . */
    , Df /* Wait for ]. */
    , Db /* Wait for . */
    , Dq /* Wait for 0 */
    , Dg /* Wait for ]0 */
    , Do /* Wait for x */
    , Dv /* Wait for 9 */
    , Dx /* Finish */
    };

#define C1 0x80000000 /* first num ready */
#define C0 0x40000000 /* single num ready */
#define Cz 0x08000000 /* new pair */
#define Cd 0x00800000 /* decimal +digit */
#define Co 0x00400000 /* octal */
#define Ch 0x00200000 /* hex */
#define Cp 0x00080000 /* +digit */
#define Cb 0x00040000 /* +hdigit */
#define C2 0x00008000 /* num pair ready */
#define Ce 0x00000100 /* Error */
#define C_state  0xFF

    static const char *clasnames[] =
    {  "L0", "Lb", "Le", "Li",  "Lq", "Ld", "Lh", "Lg"
    ,  "Lz", "Lp", "Ls", "La",  "Lx"
    };
    static const char *statenames[] =
    {  "Di", "Da", "Dd", "De",  "Dh", "Dc", "Dp", "Df"
    ,  "Db", "Dq", "Dg", "Do",  "Dv", "Dx"
    };
    static const char *pn[] = 
    {  "1 ", "0 ", "- ", "- ",  "z ", "- ", "- ", "- "
    ,  "d ", "o ", "h ", "- ",  "p ", "b ", "- ", "- "
    ,  "2 ", "- ", "- ", "- ",  "- ", "- ", "- ", "e "
    };

    static const u_int32_t automa[Dx][Lx] =
    /*             (      )      [      ]            9      x      F      0      .             ? */
    /*     L0     Lb     Le     Li     Lq           Ld     Lh     Lg     Lz     Lp     Ls     La */
    { { Ce|Dx,    Dd, Ce|Dx,    Da, Ce|Dx,       Ce|Dx, Ce|Dx, Ce|Dx, Ce|Dx, Ce|Dx,    Di, Ce|Dx}/*Di*/
    , { Ce|Dx,    Da,    Da,    Da,    Di,          Da,    Da,    Da,    Da,    Da,    Da,    Da}/*Da*/
    , { Ce|Dx, Ce|Dx,    Dx,    De, Ce|Dx, Cz|Cd|Cp|Dc, Ce|Dx, Ce|Dx, Cz|Dh, Ce|Dx,    Dd, Ce|Dx}/*Dd*/
    , { Ce|Dx,    De,    De,    De,    Dd,          De,    De,    De,    De,    De,    De,    De}/*De*/
    , { Ce|Dx, Ce|Dx, C1|Dx, C1|Df, Ce|Dx,    Co|Cp|Dc, Ch|Dc, Ce|Dx, Co|Dc, C1|Db, C1|Dp, Ce|Dx}/*Dh*/
    , { Ce|Dx, Ce|Dx, C1|Dx, C1|Df, Ce|Dx,       Cp|Dc, Ce|Dx, Cb|Dc, Cp|Dc, C1|Db, C1|Dp, Ce|Dx}/*Dc*/
    , { Ce|Dx, Ce|Dx, Ce|Dx,    Df, Ce|Dx, C0|Cd|Cp|Dc, Ce|Dx, Ce|Dx, C0|Dh,    Db,    Dp, Ce|Dx}/*Dp*/
    , { Ce|Dx,    Df,    Df,    Df,    Dp,          Df,    Df,    Df,    Df,    Df,    Df,    Df}/*Df*/
    , { Ce|Dx, Ce|Dx, Ce|Dx, Ce|Dx, Ce|Dx,       Ce|Dx, Ce|Dx, Ce|Dx, Ce|Dx,    Dq, Ce|Dx, Ce|Dx}/*Db*/
    , { Ce|Dx, Ce|Dx, C0|Dx,    Dg, Ce|Dx, Cz|Cd|Cp|Dv, Ce|Dx, Ce|Dx, Cz|Do, Ce|Dx,    Dq, Ce|Dx}/*Dq*/
    , { Ce|Dx,    Dg,    Dg,    Dg,    Dq,          Dg,    Dg,    Dg,    Dg,    Dg,    Dg,    Dg}/*Dg*/
    , { Ce|Dx, Ce|Dx, C2|Dx, C2|De, Ce|Dx,    Co|Cp|Dv, Ch|Dv, Ce|Dx, Co|Dv, Ce|Dx, C2|Dd, Ce|Dx}/*Do*/
    , { Ce|Dx, Ce|Dx, C2|Dx, C2|De, Ce|Dx,       Cp|Dv, Ce|Dx, Cb|Dv, Cp|Dv, Ce|Dx, C2|Dd, Ce|Dx}/*Dv*/
    };
    u_int32_t r0 = 0, r1 = 0, d = 0, control = 0;
    enum states state;
    int ex = EX_OK;

    ifBLIN_QV7(cnf->flags) {
         size_t i, j;
        for (j = 0; j < Lx; j++) fprintf(stderr, "      %s ", clasnames[j]);
        fprintf(stderr, "\n");
        for (i = 0; i < Dx; i++) {
            for (j = 0; j < Lx; j++) fprintf(stderr, "%08X ", automa[i][j]);
            fprintf(stderr, " %s\n", statenames[i]);
    }   }
    ifBLIN_QV5(cnf->flags) BLIN_STATEHEADER(stderr, pn);
    for (state = (2 & flocal.waitq) ? Da : (1 & flocal.waitq) ? Dd : Di; state < Dx; (*it)++) {
        control = automa[state][recode[**it]];
        ifBLIN_QV5(cnf->flags)
            BLIN_STATEBODY(stderr, pn, d, clasnames[recode[**it]], *it, 0, r0, control, statenames, state, C_state, ((*it)[BLIN_STATEVAR_i]>=' '));
        if  (C1 & control) r1 = r0;
        if  (C0 & control) SET((void*)&(cnf->conf) + acoffs[flocal.secnum], r1);
        if  ((Cz | C0 | C1) & control) r0 = 0;
        if  (Cd & control) d = 10;
        if  (Co & control) d = 8;
        if  (Ch & control) d = 16;
        if  (Cp & control) {
            u_int32_t i;

            i = (**it) - '0';
            if  (i > d) control = Ce | Dx; else r0 = r0 * d + i;
        }
        if  (Cb & control) {
            if  (16 < d) {
                control = Ce | Dx;
            } else {
                u_int32_t i;

                i = (**it) - 'A';
                if  (i > 6) i -= 'a' - 'A';
                r0 = r0 * d + i + 10;
        }   }
        if  (C2 & control) {
            u_int32_t i;

            for (i = r1; i < r0; i++) SET((void*)&(cnf->conf) + acoffs[flocal.secnum], i);
        }
        if  (Ce & control) ifBLIN_QV1(cnf->flags) warnx("bsdfig: c_lists: syntax");
        state = control & C_state;
    }
    return(ex);
}

static int
c_comm(bsdfig_conf *cnf, const u_char **it, flocal flocal) {
    size_t i;

    for (i = 0; !i && **it; (*it)++) if (clacom.retable[**it] == '\n') i = 1;
    return(EX_OK);
}

static struct {
    int     (*pr)(bsdfig_conf *cnf, const u_char **it, flocal flocal);
    flocal  fl;
} react[BSDFIG_C_C + 1] =
{ {c_error, {                         0, 0, 0}} /* BSDFIG_C_NULLQ        */
, {c_endl , {                         0, 0, 0}} /* BSDFIG_C_NULL         */
, {c_none , {                         0, 0, 0}} /* BSDFIG_C_PIPEC        */
, {c_none , {                         0, 0, 0}} /* BSDFIG_C_PIPEQ        */
, {c_none , {                         0, 0, 0}} /* BSDFIG_C_PIPE         */
, {c_none , {                         0, 0, 0}} /* BSDFIG_C_QUEUEC       */
, {c_none , {                         0, 0, 0}} /* BSDFIG_C_QUEUEQ       */
, {c_none , {                         0, 0, 0}} /* BSDFIG_C_QUEUE        */
, {c_none , {                         0, 0, 0}} /* BSDFIG_C_DISABLEC     */
, {c_none , {                         0, 0, 0}} /* BSDFIG_C_DISABLEQ     */
, {c_none , {                         0, 0, 0}} /* BSDFIG_C_DISABLE      */
, {c_none , {                         0, 0, 0}} /* BSDFIG_C_RULEC        */
, {c_none , {                         0, 0, 0}} /* BSDFIG_C_RULEQ        */
, {c_none , {                         0, 0, 0}} /* BSDFIG_C_RULE         */
, {c_none , {                         0, 0, 0}} /* BSDFIG_C_TABLEC       */
, {c_none , {                         0, 0, 0}} /* BSDFIG_C_TABLEQ       */
, {c_none , {                         0, 0, 0}} /* BSDFIG_C_TABLE        */
, {c_lists, {BSDFIG_IPFW_OTHERS_MAX_S  , 0, 2}} /* BSDFIG_C_PIPELISTC    */
, {c_lists, {BSDFIG_IPFW_OTHERS_MAX_S  , 0, 1}} /* BSDFIG_C_PIPELISTQ    */
, {c_lists, {BSDFIG_IPFW_OTHERS_MAX_S  , 0, 0}} /* BSDFIG_C_PIPELIST     */
, {c_lists, {BSDFIG_IPFW_OTHERS_MAX_S  , 1, 2}} /* BSDFIG_C_QUEUELISTC   */
, {c_lists, {BSDFIG_IPFW_OTHERS_MAX_S  , 1, 1}} /* BSDFIG_C_QUEUELISTQ   */
, {c_lists, {BSDFIG_IPFW_OTHERS_MAX_S  , 1, 0}} /* BSDFIG_C_QUEUELIST    */
, {c_lists, {BSDFIG_IPFW_DISABLES_MAX_S, 2, 2}} /* BSDFIG_C_DISABLELISTC */
, {c_lists, {BSDFIG_IPFW_DISABLES_MAX_S, 2, 1}} /* BSDFIG_C_DISABLELISTQ */
, {c_lists, {BSDFIG_IPFW_DISABLES_MAX_S, 2, 0}} /* BSDFIG_C_DISABLELIST  */
, {c_lists, {BSDFIG_IPFW_OTHERS_MAX_S  , 3, 2}} /* BSDFIG_C_RULELISTC    */
, {c_lists, {BSDFIG_IPFW_OTHERS_MAX_S  , 3, 1}} /* BSDFIG_C_RULELISTQ    */
, {c_lists, {BSDFIG_IPFW_OTHERS_MAX_S  , 3, 0}} /* BSDFIG_C_RULELIST     */
, {c_lists, {BSDFIG_IPFW_TABLES_MAX_S  , 4, 2}} /* BSDFIG_C_TABLELISTC   */
, {c_lists, {BSDFIG_IPFW_TABLES_MAX_S  , 4, 1}} /* BSDFIG_C_TABLELISTQ   */
, {c_lists, {BSDFIG_IPFW_TABLES_MAX_S  , 4, 0}} /* BSDFIG_C_TABLELIST    */
, {c_comm , {                         0, 0, 0}} /* BSDFIG_C_C            */
};

int
bsdfig_confread(bsdfig_conf *cnf, mife_descriptor *in) {
    enum bsdfig_section sec;
    const u_char *ipt;
    size_t sztext, pt;
    int ex = EX_OK;
    u_char *intext;
    off_t ontext;
    unsigned i;

    if  (!cnf) errx(EX_USAGE, "bsdfig_confread: no cfg");
    ifBLIN_QV4(cnf->flags) fprintf(stderr, "+bsdfig_confread %08X\n", cnf->flags);
    if  (mife_eolinit(in, "\n") < 0) {
        ifBLIN_QV1(cnf->flags) warn("bsdfig_confread: mife_eolinit");
        ex = EX_SOFTWARE;
    }
    for (ontext = 0; !ex && !(MIFE_EOFL & in->flag); ontext += pt) {
        if  ((int)(sztext = mife_read(in, 0, ontext)) < 0) {
            ifBLIN_QV1(cnf->flags) warn("bsdfig_confread: mife_read");
            ex = EX_IOERR;
            break;
        }
        pt = 0;
        if  (!sztext) {
            ifBLIN_QV6(cnf->flags) fprintf(stderr, "0[%"BLIN_U"] End\n", pt);
            break;
        }
        if  (!(intext = mife_get(in, ontext))) {
            ifBLIN_QV1(cnf->flags) warn("bsdfig_confread: mife_get");
            ex = EX_IOERR;
            break;
        }
        if  (!intext[pt]) {
            ifBLIN_QV6(cnf->flags) fprintf(stderr, "1[%"BLIN_U"] End\n", pt);
            break;
        }
        ifBLIN_QV6(cnf->flags) {
            fprintf(stderr, "2[%"BLIN_U"]=", pt);
            for (i = 0; (i < sztext) && (intext[i] >= ' '); i++) fprintf(stderr, "%c", intext[i]);
            fprintf(stderr, "~%02X\n", intext[i]);
        }
        for (pt = 0; (pt < sztext) && (clacom.retable[intext[pt]] == ' '); pt++);
        ifBLIN_QV6(cnf->flags) {
            fprintf(stderr, "3[%"BLIN_U"]=", pt);
            for (i = pt; (i < sztext) && (intext[i] >= ' '); i++) fprintf(stderr, "%c", intext[i]);
            fprintf(stderr, "~%02X\n", intext[i]);
        }
        ipt = &(intext[pt]);
        sec = babolo_goword(&clacom, &ipt);
        if  (sec > BSDFIG_C_C) sec = BSDFIG_C_C;
        pt += ipt - &(intext[pt]);
        ifBLIN_QV6(cnf->flags) {
            fprintf(stderr, "4[%"BLIN_U"]=", pt);
            for (i = pt; (i < sztext) && (intext[i] >= ' '); i++) fprintf(stderr, "%c", intext[i]);
            fprintf(stderr, "~%02X\n", intext[i]);
        }
        ex = react[sec].pr(cnf, &ipt, react[sec].fl); /* XXXX */
        pt += ipt - &(intext[pt]);
        ifBLIN_QV6(cnf->flags) {
            fprintf(stderr, "5[%"BLIN_U"]=", pt);
            for (i = pt; (i < sztext) && (intext[i] >= ' '); i++) fprintf(stderr, "%c", intext[i]);
            fprintf(stderr, "~%02X\n", intext[i]);
        }
        for (i = 0; !i && intext[pt];) if (clacom.retable[intext[pt++]] == '\n') i = 1;
        ifBLIN_QV6(cnf->flags) {
            fprintf(stderr, "6[%"BLIN_U"]=", pt);
            for (i = pt; (i < sztext) && (intext[i] >= ' '); i++) fprintf(stderr, "%c", intext[i]);
            fprintf(stderr, "~%02X\n", intext[i]);
        }
    }
    ifBLIN_QV4(cnf->flags) fprintf(stderr, "-bsdfig_confread %d\n", ex);
    return(ex);
}
